@@ -17,18 +17,25 @@
 package org.gradle.kotlin.dsl.accessors
 
 import org.gradle.api.Project
+import org.gradle.api.initialization.Settings
+import org.gradle.api.internal.GradleInternal
+import org.gradle.api.internal.SettingsInternal
 import org.gradle.api.internal.file.FileCollectionFactory
+import org.gradle.api.internal.initialization.ClassLoaderScope
 import org.gradle.api.internal.project.ProjectInternal
-import org.gradle.api.tasks.ClasspathNormalizer
-import org.gradle.internal.classanalysis.AsmConstants.ASM_LEVEL
+import org.gradle.api.plugins.ExtensionAware
+import org.gradle.initialization.GradlePropertiesController
+import org.gradle.internal.classloader.ClassLoaderUtils
 import org.gradle.internal.classpath.ClassPath
 import org.gradle.internal.classpath.DefaultClassPath
 import org.gradle.internal.execution.ExecutionEngine
+import org.gradle.internal.execution.ImmutableUnitOfWork
+import org.gradle.internal.execution.InputFingerprinter
 import org.gradle.internal.execution.UnitOfWork
-import org.gradle.internal.execution.fingerprint.InputFingerprinter
-import org.gradle.internal.execution.fingerprint.InputFingerprinter.FileValueSupplier
-import org.gradle.internal.execution.fingerprint.InputFingerprinter.InputPropertyType.NON_INCREMENTAL
-import org.gradle.internal.execution.fingerprint.InputFingerprinter.InputVisitor
+import org.gradle.internal.execution.UnitOfWork.InputFileValueSupplier
+import org.gradle.internal.execution.UnitOfWork.InputVisitor
+import org.gradle.internal.execution.UnitOfWork.OutputFileValueSupplier
+import org.gradle.internal.execution.model.InputNormalizer
 import org.gradle.internal.file.TreeType.DIRECTORY
 import org.gradle.internal.fingerprint.CurrentFileCollectionFingerprint
 import org.gradle.internal.fingerprint.DirectorySensitivity
@@ -36,15 +43,23 @@ import org.gradle.internal.fingerprint.LineEndingSensitivity
 import org.gradle.internal.hash.HashCode
 import org.gradle.internal.hash.Hasher
 import org.gradle.internal.hash.Hashing
+import org.gradle.internal.properties.InputBehavior.NON_INCREMENTAL
+import org.gradle.internal.service.scopes.Scope
+import org.gradle.internal.service.scopes.ServiceScope
 import org.gradle.internal.snapshot.ValueSnapshot
 import org.gradle.kotlin.dsl.cache.KotlinDslWorkspaceProvider
-import org.gradle.kotlin.dsl.codegen.fileHeaderFor
-import org.gradle.kotlin.dsl.codegen.kotlinDslPackageName
+import org.gradle.kotlin.dsl.concurrent.AsyncIOScopeFactory
 import org.gradle.kotlin.dsl.concurrent.IO
-import org.gradle.kotlin.dsl.concurrent.withAsynchronousIO
-import org.gradle.kotlin.dsl.support.ClassBytesRepository
-import org.gradle.kotlin.dsl.support.appendReproducibleNewLine
+import org.gradle.kotlin.dsl.concurrent.runBlocking
+import org.gradle.kotlin.dsl.internal.sharedruntime.codegen.KOTLIN_DSL_PACKAGE_NAME
+import org.gradle.kotlin.dsl.internal.sharedruntime.codegen.fileHeaderFor
+import org.gradle.kotlin.dsl.internal.sharedruntime.codegen.primitiveKotlinTypeNames
+import org.gradle.kotlin.dsl.internal.sharedruntime.support.ClassBytesRepository
+import org.gradle.kotlin.dsl.internal.sharedruntime.support.appendReproducibleNewLine
+import org.gradle.kotlin.dsl.support.getBooleanKotlinDslOption
+import org.gradle.kotlin.dsl.support.serviceOf
 import org.gradle.kotlin.dsl.support.useToRun
+import org.gradle.model.internal.asm.AsmConstants.ASM_LEVEL
 import org.jetbrains.kotlin.metadata.ProtoBuf
 import org.jetbrains.kotlin.metadata.ProtoBuf.Visibility
 import org.jetbrains.kotlin.metadata.deserialization.Flags
@@ -61,60 +76,81 @@ import java.io.File
 import javax.inject.Inject
 
 
-class ProjectAccessorsClassPathGenerator @Inject constructor(
+@ServiceScope(Scope.Build::class)
+class ProjectAccessorsClassPathGenerator @Inject internal constructor(
     private val fileCollectionFactory: FileCollectionFactory,
     private val projectSchemaProvider: ProjectSchemaProvider,
     private val executionEngine: ExecutionEngine,
     private val inputFingerprinter: InputFingerprinter,
-    private val workspaceProvider: KotlinDslWorkspaceProvider
+    private val workspaceProvider: KotlinDslWorkspaceProvider,
+    private val asyncIO: AsyncIOScopeFactory,
 ) {
 
-    fun projectAccessorsClassPath(project: Project, classPath: ClassPath): AccessorsClassPath =
-        project.getOrCreateProperty("gradleKotlinDsl.projectAccessorsClassPath") {
-            buildAccessorsClassPathFor(project, classPath)
+    fun projectAccessorsClassPath(scriptTarget: ExtensionAware, classPath: ClassPath): AccessorsClassPath =
+        scriptTarget.getOrCreateProperty("gradleKotlinDsl.accessorsClassPath") {
+            buildAccessorsClassPathFor(scriptTarget, classPath)
                 ?: AccessorsClassPath.empty
         }
 
 
     private
-    fun buildAccessorsClassPathFor(project: Project, classPath: ClassPath): AccessorsClassPath? {
-        return configuredProjectSchemaOf(project)?.let { projectSchema ->
-            val work = GenerateProjectAccessors(
-                project,
-                projectSchema,
-                classPath,
-                fileCollectionFactory,
-                inputFingerprinter,
-                workspaceProvider
-            )
-            val result = executionEngine.createRequest(work).execute()
-            result.executionResult.get().output as AccessorsClassPath
-        }
-    }
+    fun buildAccessorsClassPathFor(scriptTarget: Any, classPath: ClassPath): AccessorsClassPath? =
+        classLoaderScopeOf(scriptTarget)
+            ?.let { classLoaderScope ->
+                configuredProjectSchemaOf(scriptTarget, classLoaderScope)
+            }?.let { scriptTargetSchema ->
+                val work = GenerateProjectAccessors(
+                    scriptTarget,
+                    scriptTargetSchema,
+                    classPath,
+                    fileCollectionFactory,
+                    inputFingerprinter,
+                    workspaceProvider,
+                    asyncIO,
+                    isDclEnabledForScriptTarget(scriptTarget),
+                )
+                executionEngine.createRequest(work)
+                    .execute()
+                    .getOutputAs(AccessorsClassPath::class.java)
+                    .get()
+            }
 
 
     private
-    fun configuredProjectSchemaOf(project: Project): TypedProjectSchema? =
-        if (enabledJitAccessors(project)) {
-            require(classLoaderScopeOf(project).isLocked) {
-                "project.classLoaderScope must be locked before querying the project schema"
-            }
-            projectSchemaProvider.schemaFor(project).takeIf { it.isNotEmpty() }
-        } else null
+    fun configuredProjectSchemaOf(scriptTarget: Any, classLoaderScope: ClassLoaderScope): TypedProjectSchema? {
+        require(classLoaderScope.isLocked) {
+            "project.classLoaderScope must be locked before querying the project schema"
+        }
+        return projectSchemaProvider.schemaFor(scriptTarget, classLoaderScope)?.takeIf { it.isNotEmpty() }
+    }
 }
 
+fun isDclEnabledForScriptTarget(target: Any): Boolean {
+    val gradleProperties = when (target) {
+        is Project -> target.serviceOf<GradlePropertiesController>()
+        is Settings -> target.serviceOf<GradlePropertiesController>()
+        else -> null
+    }
+    return gradleProperties?.let { getBooleanKotlinDslOption(it, DCL_ENABLED_PROPERTY_NAME, false) } ?: false
+}
 
+const val DCL_ENABLED_PROPERTY_NAME = "org.gradle.kotlin.dsl.dcl"
+
+internal
 class GenerateProjectAccessors(
-    private val project: Project,
-    private val projectSchema: TypedProjectSchema,
+    private val scriptTarget: Any,
+    private val scriptTargetSchema: TypedProjectSchema,
     private val classPath: ClassPath,
     private val fileCollectionFactory: FileCollectionFactory,
     private val inputFingerprinter: InputFingerprinter,
-    private val workspaceProvider: KotlinDslWorkspaceProvider
-) : UnitOfWork {
+    private val workspaceProvider: KotlinDslWorkspaceProvider,
+    private val asyncIO: AsyncIOScopeFactory,
+    private val isDclEnabled: Boolean
+) : ImmutableUnitOfWork {
 
     companion object {
-        const val PROJECT_SCHEMA_INPUT_PROPERTY = "projectSchema"
+        const val TARGET_SCHEMA_INPUT_PROPERTY = "targetSchema"
+        const val DCL_ENABLED_INPUT_PROPERTY = "dcl"
         const val CLASSPATH_INPUT_PROPERTY = "classpath"
         const val SOURCES_OUTPUT_PROPERTY = "sources"
         const val CLASSES_OUTPUT_PROPERTY = "classes"
@@ -122,9 +158,9 @@ class GenerateProjectAccessors(
 
     override fun execute(executionRequest: UnitOfWork.ExecutionRequest): UnitOfWork.WorkOutput {
         val workspace = executionRequest.workspace
-        withAsynchronousIO(project) {
+        asyncIO.runBlocking {
             buildAccessorsFor(
-                projectSchema,
+                scriptTargetSchema,
                 classPath,
                 srcDir = getSourcesOutputDir(workspace),
                 binDir = getClassesOutputDir(workspace)
@@ -133,18 +169,19 @@ class GenerateProjectAccessors(
         return object : UnitOfWork.WorkOutput {
             override fun getDidWork() = UnitOfWork.WorkResult.DID_WORK
 
-            override fun getOutput() = loadRestoredOutput(workspace)
+            override fun getOutput(workspace: File) = loadAlreadyProducedOutput(workspace)
         }
     }
 
-    override fun loadRestoredOutput(workspace: File) = AccessorsClassPath(
+    override fun loadAlreadyProducedOutput(workspace: File) = AccessorsClassPath(
         DefaultClassPath.of(getClassesOutputDir(workspace)),
         DefaultClassPath.of(getSourcesOutputDir(workspace))
     )
 
     override fun identify(identityInputs: Map<String, ValueSnapshot>, identityFileInputs: Map<String, CurrentFileCollectionFingerprint>): UnitOfWork.Identity {
         val hasher = Hashing.newHasher()
-        requireNotNull(identityInputs[PROJECT_SCHEMA_INPUT_PROPERTY]).appendToHasher(hasher)
+        requireNotNull(identityInputs[TARGET_SCHEMA_INPUT_PROPERTY]).appendToHasher(hasher)
+        requireNotNull(identityInputs[DCL_ENABLED_INPUT_PROPERTY]).appendToHasher(hasher)
         hasher.putHash(requireNotNull(identityFileInputs[CLASSPATH_INPUT_PROPERTY]).hash)
         val identityHash = hasher.hash().toString()
         return UnitOfWork.Identity { identityHash }
@@ -154,16 +191,17 @@ class GenerateProjectAccessors(
 
     override fun getInputFingerprinter() = inputFingerprinter
 
-    override fun getDisplayName(): String = "Kotlin DSL accessors for $project"
+    override fun getDisplayName(): String = "Kotlin DSL accessors for $scriptTarget"
 
     override fun visitIdentityInputs(visitor: InputVisitor) {
-        visitor.visitInputProperty(PROJECT_SCHEMA_INPUT_PROPERTY) { hashCodeFor(projectSchema) }
+        visitor.visitInputProperty(TARGET_SCHEMA_INPUT_PROPERTY) { hashCodeFor(scriptTargetSchema) }
+        visitor.visitInputProperty(DCL_ENABLED_INPUT_PROPERTY) { isDclEnabled }
         visitor.visitInputFileProperty(
             CLASSPATH_INPUT_PROPERTY,
             NON_INCREMENTAL,
-            FileValueSupplier(
+            InputFileValueSupplier(
                 classPath,
-                ClasspathNormalizer::class.java,
+                InputNormalizer.RUNTIME_CLASSPATH,
                 DirectorySensitivity.IGNORE_DIRECTORIES,
                 LineEndingSensitivity.DEFAULT,
             ) { fileCollectionFactory.fixed(classPath.asFiles) }
@@ -173,8 +211,8 @@ class GenerateProjectAccessors(
     override fun visitOutputs(workspace: File, visitor: UnitOfWork.OutputVisitor) {
         val sourcesOutputDir = getSourcesOutputDir(workspace)
         val classesOutputDir = getClassesOutputDir(workspace)
-        visitor.visitOutputProperty(SOURCES_OUTPUT_PROPERTY, DIRECTORY, sourcesOutputDir, fileCollectionFactory.fixed(sourcesOutputDir))
-        visitor.visitOutputProperty(CLASSES_OUTPUT_PROPERTY, DIRECTORY, classesOutputDir, fileCollectionFactory.fixed(classesOutputDir))
+        visitor.visitOutputProperty(SOURCES_OUTPUT_PROPERTY, DIRECTORY, OutputFileValueSupplier.fromStatic(sourcesOutputDir, fileCollectionFactory.fixed(sourcesOutputDir)))
+        visitor.visitOutputProperty(CLASSES_OUTPUT_PROPERTY, DIRECTORY, OutputFileValueSupplier.fromStatic(classesOutputDir, fileCollectionFactory.fixed(classesOutputDir)))
     }
 }
 
@@ -203,7 +241,7 @@ fun IO.buildAccessorsFor(
     classPath: ClassPath,
     srcDir: File,
     binDir: File?,
-    packageName: String = kotlinDslPackageName,
+    packageName: String = KOTLIN_DSL_PACKAGE_NAME,
     format: AccessorFormat = AccessorFormats.default
 ) {
     val availableSchema = availableProjectSchemaFor(projectSchema, classPath)
@@ -226,6 +264,7 @@ object AccessorFormats {
         accessor.trimIndent()
     }
 
+    @Suppress("ObjectPropertyNaming")
     val `internal`: AccessorFormat = { accessor ->
         accessor
             .trimIndent()
@@ -265,8 +304,10 @@ fun availableProjectSchemaFor(projectSchema: TypedProjectSchema, classPath: Clas
 
 
 sealed class TypeAccessibility {
-    data class Accessible(val type: SchemaType) : TypeAccessibility()
-    data class Inaccessible(val type: SchemaType, val reasons: List<InaccessibilityReason>) : TypeAccessibility()
+    abstract val type: SchemaType
+
+    data class Accessible(override val type: SchemaType) : TypeAccessibility()
+    data class Inaccessible(override val type: SchemaType, val reasons: List<InaccessibilityReason>) : TypeAccessibility()
 }
 
 
@@ -298,7 +339,10 @@ internal
 class TypeAccessibilityProvider(classPath: ClassPath) : Closeable {
 
     private
-    val classBytesRepository = ClassBytesRepository(classPath)
+    val classBytesRepository = ClassBytesRepository(
+        ClassLoaderUtils.getPlatformClassLoader(),
+        classPath.asFiles
+    )
 
     private
     val typeAccessibilityInfoPerClass = mutableMapOf<String, TypeAccessibilityInfo>()
@@ -388,7 +432,6 @@ fun classNamesFromTypeString(type: SchemaType): ClassNamesFromTypeString =
 
 internal
 fun classNamesFromTypeString(typeString: String): ClassNamesFromTypeString {
-
     val all = mutableListOf<String>()
     val leafs = mutableListOf<String>()
     var buffer = StringBuilder()
@@ -405,6 +448,7 @@ fun classNamesFromTypeString(typeString: String): ClassNamesFromTypeString {
                 nonPrimitiveKotlinType()?.also { all.add(it) }
                 buffer = StringBuilder()
             }
+
             in " ,>" -> {
                 nonPrimitiveKotlinType()?.also {
                     all.add(it)
@@ -412,6 +456,7 @@ fun classNamesFromTypeString(typeString: String): ClassNamesFromTypeString {
                 }
                 buffer = StringBuilder()
             }
+
             else -> buffer.append(char)
         }
     }
@@ -450,6 +495,7 @@ class KotlinVisibilityClassVisitor : ClassVisitor(ASM_LEVEL) {
             "Lkotlin/Metadata;" -> ClassDataFromKotlinMetadataAnnotationVisitor { classData ->
                 visibility = Flags.VISIBILITY[classData.flags]
             }
+
             else -> null
         }
 }
@@ -499,7 +545,7 @@ class AnnotationValueCollector<T>(val output: MutableList<T>) : AnnotationVisito
 }
 
 
-internal
+private
 operator fun Int.contains(flag: Int) =
     and(flag) == flag
 
@@ -540,8 +586,12 @@ fun inaccessible(type: SchemaType, reasons: List<InaccessibilityReason>): TypeAc
 
 
 private
-fun classLoaderScopeOf(project: Project) =
-    (project as ProjectInternal).classLoaderScope
+fun classLoaderScopeOf(scriptTarget: Any) = when (scriptTarget) {
+    is ProjectInternal -> scriptTarget.classLoaderScope
+    is SettingsInternal -> scriptTarget.classLoaderScope
+    is GradleInternal -> scriptTarget.classLoaderScope
+    else -> null
+}
 
 
 fun hashCodeFor(schema: TypedProjectSchema): HashCode = Hashing.newHasher().run {
@@ -549,6 +599,8 @@ fun hashCodeFor(schema: TypedProjectSchema): HashCode = Hashing.newHasher().run
     putAll(schema.conventions)
     putAll(schema.tasks)
     putAll(schema.containerElements)
+    putContainerElementFactoryEntries(schema.containerElementFactories)
+    putSoftwareTypeEntries(schema.softwareTypeEntries)
     putAllSorted(schema.configurations.map { it.target })
     hash()
 }
@@ -571,12 +623,22 @@ fun Hasher.putAll(entries: List<ProjectSchemaEntry<SchemaType>>) {
     }
 }
 
+private fun Hasher.putContainerElementFactoryEntries(entries: List<ContainerElementFactoryEntry<SchemaType>>) {
+    putInt(entries.size)
+    entries.forEach { entry ->
+        putString(entry.factoryName)
+        putString(entry.containerReceiverType.kotlinString)
+        putString(entry.publicType.kotlinString)
+    }
+}
 
-private
-fun enabledJitAccessors(project: Project) =
-    project.findProperty("org.gradle.kotlin.dsl.accessors")?.let {
-        it != "false" && it != "off"
-    } ?: true
+private fun Hasher.putSoftwareTypeEntries(entries: List<SoftwareTypeEntry<SchemaType>>) {
+    putInt(entries.size)
+    entries.forEach { entry ->
+        putString(entry.softwareTypeName)
+        putString(entry.modelType.kotlinString)
+    }
+}
 
 
 internal
@@ -584,7 +646,7 @@ fun IO.writeAccessorsTo(
     outputFile: File,
     accessors: Iterable<String>,
     imports: List<String> = emptyList(),
-    packageName: String = kotlinDslPackageName
+    packageName: String = KOTLIN_DSL_PACKAGE_NAME
 ) = io {
     outputFile.bufferedWriter().useToRun {
         appendReproducibleNewLine(fileHeaderWithImportsFor(packageName))
@@ -603,7 +665,7 @@ fun IO.writeAccessorsTo(
 
 
 internal
-fun fileHeaderWithImportsFor(accessorsPackage: String = kotlinDslPackageName) = """
+fun fileHeaderWithImportsFor(accessorsPackage: String = KOTLIN_DSL_PACKAGE_NAME) = """
 ${fileHeaderFor(accessorsPackage)}
 
 import org.gradle.api.Action
@@ -622,7 +684,9 @@ import org.gradle.api.artifacts.PublishArtifact
 import org.gradle.api.artifacts.dsl.ArtifactHandler
 import org.gradle.api.artifacts.dsl.DependencyConstraintHandler
 import org.gradle.api.artifacts.dsl.DependencyHandler
+import org.gradle.api.initialization.SharedModelDefaults
 import org.gradle.api.provider.Provider
+import org.gradle.api.provider.ProviderConvertible
 import org.gradle.api.tasks.TaskContainer
 import org.gradle.api.tasks.TaskProvider
 
@@ -635,16 +699,18 @@ import org.gradle.kotlin.dsl.accessors.runtime.*
 /**
  * Location of the discontinued project schema snapshot, relative to the root project.
  */
-const val projectSchemaResourcePath =
+internal
+const val PROJECT_SCHEMA_RESOURCE_PATH =
     "gradle/project-schema.json"
 
 
-const val projectSchemaResourceDiscontinuedWarning =
-    "Support for $projectSchemaResourcePath was removed in Gradle 5.0. The file is no longer used and it can be safely deleted."
+internal
+const val PROJECT_SCHEMA_RESOURCE_DISCONTINUED_WARNING =
+    "Support for $PROJECT_SCHEMA_RESOURCE_PATH was removed in Gradle 5.0. The file is no longer used and it can be safely deleted."
 
 
 fun Project.warnAboutDiscontinuedJsonProjectSchema() {
-    if (file(projectSchemaResourcePath).isFile) {
-        logger.warn(projectSchemaResourceDiscontinuedWarning)
+    if (file(PROJECT_SCHEMA_RESOURCE_PATH).isFile) {
+        logger.warn(PROJECT_SCHEMA_RESOURCE_DISCONTINUED_WARNING)
     }
 }
